# SPDX-FileCopyrightText: 2025-present Doug Richardson <git@rekt.email>
# SPDX-License-Identifier: MIT

import webbrowser

import click

from dogcrud.core.context import config_context
from dogcrud.core.data import resource_type_for_filename


@click.command(name="open")
@click.argument("filename", type=click.Path(exists=True))
def open_in_browser(filename: str) -> None:
    """
    Open Datadog web page corresponding to FILENAME.
    """
    config_context().run_in_context(async_open(filename))


async def async_open(filename: str) -> None:
    rt = resource_type_for_filename(filename)
    resource_id = rt.resource_id(filename)
    webbrowser.open_new_tab(rt.webpage_url(resource_id))



# SPDX-FileCopyrightText: 2025-present Doug Richardson <git@rekt.email>
# SPDX-License-Identifier: MIT

import asyncio
import logging
from collections.abc import AsyncGenerator
from pathlib import Path
from typing import TYPE_CHECKING, override

import aiofiles
from pydantic import BaseModel

from dogcrud.core import context, rest
from dogcrud.core.metrics import list_metrics
from dogcrud.core.resource_type import IDType, ResourceType

if TYPE_CHECKING:
    from dogcrud.core.pagination import CursorDataItemModel

logger = logging.getLogger(__name__)


class MetricTagAggregationModel(BaseModel):
    space: str
    time: str


class MetricTagAttributesModel(BaseModel):
    aggregations: list[MetricTagAggregationModel] | None = None
    exclude_tags_mode: bool | None = None
    include_percentiles: bool | None = None
    metric_type: str | None
    tags: list[str] | None = None


class MetricTagDataModel(BaseModel):
    attributes: MetricTagAttributesModel | None = None
    id: str
    type: str


class MetricTagModel(BaseModel):
    data: MetricTagDataModel


class MetricResourceType(ResourceType):
    """
    A Datadog metric resource. This resource doesn't fit in with a StandardResourceType since
    - you have an idea of "active" metrics that the initial list is populated from from a time range
    - you get a list of metrics with a subset of metadata
    - there is no get ID list and get specific resource json like StandardResourceType
    """

    def __init__(
        self,
        max_concurrency: int,
        *,
        disabled: bool = False,
    ) -> None:
        self.rest_base_path = "v2/metrics"
        self.concurrency_semaphore = asyncio.BoundedSemaphore(max_concurrency)
        self._id_to_item_index: dict[IDType, CursorDataItemModel] = {}
        self.disabled = disabled

    @override
    def rest_path(self, resource_id: IDType | None = None) -> str:
        match resource_id:
            case None:
                return self.rest_base_path
            case _:
                return f"{self.rest_base_path}/{resource_id}"

    @override
    def local_path(self, resource_id: IDType | None = None) -> Path:
        data_dir = context.config_context().data_dir
        match resource_id:
            case None:
                return data_dir / self.rest_base_path
            case _:
                return data_dir / f"{self.rest_base_path}/{resource_id}.json"

    @override
    async def get(self, resource_id: IDType) -> bytes:
        # TODO: load _id_to_index_item if not yet set
        return self._id_to_item_index[resource_id].json().encode("utf-8")

    @override
    async def put(self, resource_id: IDType, data: bytes) -> None:
        url = f"api/v2/metrics/{resource_id}/tags"
        async with self.concurrency_semaphore:
            await rest.patch_json(url, data)

    @override
    def transform_get_to_put(self, data: bytes) -> bytes:
        metric_tag_data = MetricTagDataModel.model_validate_json(data)
        metric_tag = MetricTagModel(data=metric_tag_data)
        return metric_tag.model_dump_json(exclude_none=True).encode()

    @override
    async def list_ids(self) -> AsyncGenerator[IDType]:
        async for metric in list_metrics(self.concurrency_semaphore):
            self._id_to_item_index[metric.id] = metric
            yield metric.id

    @override
    async def read_local_json(self, resource_id: IDType) -> bytes:
        resource_dir = context.config_context().data_dir / self.rest_path(resource_id)
        filename = str(resource_dir / f"{resource_id}.json")

        async with aiofiles.open(filename, "rb") as file:
            return await file.read()

    @override
    def webpage_url(self, resource_id: IDType) -> str:
        return f"https://app.datadoghq.com/metric/summary?metric={resource_id}"

    @override
    def resource_id(self, filename: str) -> IDType:
        return Path(filename).stem



# SPDX-FileCopyrightText: 2025-present Doug Richardson <git@rekt.email>
# SPDX-License-Identifier: MIT

import asyncio
import logging
from collections.abc import AsyncGenerator, AsyncIterator, Sequence
from dataclasses import dataclass
from typing import Any, Protocol, assert_never, override
from urllib.parse import urlencode

import orjson
from pydantic import BaseModel, ConfigDict

from dogcrud.core.resource_type import IDType
from dogcrud.core.rest import get_json

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Page:
    ids: Sequence[IDType]
    items: Sequence[Any]
    parsed_json: Any


class PaginationStrategy(Protocol):
    def pages(self, url: str, concurrency_semaphore: asyncio.Semaphore) -> AsyncIterator[Page]: ...


async def _get_page(url: str, items_key: str | None) -> Page:
    json = await get_json(url)
    parsed_json = orjson.loads(json)

    match (parsed_json, items_key):
        case list(), None:
            items = parsed_json
        case dict(), items_key if items_key is not None:
            items = parsed_json[items_key]
        case _, _:
            msg = f"Invalid combination of page response type {type(parsed_json)} and items key '{items_key}'"
            raise RuntimeError(msg)

    ids = [item["id"] for item in items]
    return Page(ids=ids, items=items, parsed_json=parsed_json)


@dataclass(frozen=True)
class ItemOffsetPagination(PaginationStrategy):
    """
    A pagination strategy based on offsets.

    Attributes:
        offset_query_param: The name of the query parameter for the offset.
        items_key: If the page response is an object, this is the key to the list of items. Use None if the response is already a list.
    """

    offset_query_param: str
    items_key: str | None = None

    @override
    async def pages(self, url: str, concurrency_semaphore: asyncio.Semaphore) -> AsyncIterator[Page]:
        offset = 0
        seen_ids: set[IDType] = set()
        while True:
            async with concurrency_semaphore:
                page = await _get_page(f"{url}?{self.offset_query_param}={offset}", self.items_key)
            if not page.ids:
                return
            yield page
            offset += len(page.ids)
            new_ids = set(page.ids)
            if already_seen_ids := new_ids & seen_ids:
                logger.warning(
                    f"Duplicate IDs seen while paging. If transient okay to ignore cause it might be someone else mutating the list of items. already_seen_ids={already_seen_ids}. If not transient, this could be a misconfiguration of the pagination_strategy."
                )
            seen_ids |= new_ids


@dataclass(frozen=True)
class LimitOffsetPagination(PaginationStrategy):
    """
    A pagination strategy based on limit and offset query parameters.

    Attributes:
        limit: The number of items to fetch per page.
        limit_query_param: The name of the query parameter for the limit.
        offset_query_param: The name of the query parameter for the offset.
        items_key: If the page response is an object, this is the key to the list of items. Use None if the response is already a list.
    """

    limit: int = 100
    limit_query_param: str = "limit"
    offset_query_param: str = "offset"
    items_key: str | None = None

    @override
    async def pages(self, url: str, concurrency_semaphore: asyncio.Semaphore) -> AsyncIterator[Page]:
        offset = 0
        seen_ids: set[IDType] = set()
        while True:
            async with concurrency_semaphore:
                query_params = {
                    self.limit_query_param: self.limit,
                    self.offset_query_param: offset,
                }
                page = await _get_page(
                    f"{url}?{urlencode(query_params)}",
                    self.items_key,
                )
            if not page.ids:
                return
            yield page
            offset += len(page.ids)
            new_ids = set(page.ids)
            if already_seen_ids := new_ids & seen_ids:
                logger.warning(
                    f"Duplicate IDs seen while paging. If transient okay to ignore cause it might be someone else mutating the list of items. already_seen_ids={already_seen_ids}. If not transient, this could be a misconfiguration of the pagination_strategy."
                )
            seen_ids |= new_ids


@dataclass(frozen=True)
class IDOffsetPagination(PaginationStrategy):
    """
    A pagination strategy based on last seen ID.

    Attributes:
        offset_query_param: The name of the query parameter for the offset.
        items_key: If the page response is an object, this is the key to the list of items. Use None if the response is already a list.
    """

    offset_query_param: str
    items_key: str | None = None

    @override
    async def pages(self, url: str, concurrency_semaphore: asyncio.Semaphore) -> AsyncIterator[Page]:
        id_offset: IDType = "0"
        offset = 0
        seen_ids: set[IDType] = set()
        while True:
            async with concurrency_semaphore:
                page = await _get_page(f"{url}?{self.offset_query_param}={id_offset}", self.items_key)
            if not page.ids:
                return
            yield page
            id_offset = page.ids[-1]
            offset += len(page.ids)
            new_ids = set(page.ids)
            if already_seen_ids := new_ids & seen_ids:
                logger.warning(
                    f"Duplicate IDs seen while paging. If transient okay to ignore cause it might be someone else mutating the list of items. already_seen_ids={already_seen_ids}. If not transient, this could be a misconfiguration of the pagination_strategy."
                )
            seen_ids |= new_ids


@dataclass(frozen=True)
class NoPagination(PaginationStrategy):
    """
    A pagination strategy that fetches everything in one go.

    Attributes:
        items_key: If the page response is an object, this is the key to the list of items. Use None if the response is already a list.
    """

    items_key: str | None = None

    @override
    async def pages(self, url: str, concurrency_semaphore: asyncio.Semaphore) -> AsyncIterator[Page]:
        async with concurrency_semaphore:
            page = await _get_page(url, self.items_key)
        yield page


class CursorPaginationModel(BaseModel):
    next_cursor: str | None


class CursorMetadataModel(BaseModel):
    pagination: CursorPaginationModel


class CursorLinksModel(BaseModel):
    first: str | None
    last: str | None
    next: str | None
    prev: str | None
    self: str | None


class CursorDataItemModel(BaseModel):
    model_config = ConfigDict(extra="allow")

    id: str


class CursorPageModel(BaseModel):
    data: list[CursorDataItemModel]
    links: CursorLinksModel
    meta: CursorMetadataModel


@dataclass(frozen=True)
class CursorPagination:
    """
    A pagination strategy based on last seen page cursor.
    """

    query_params: str = ""

    async def pages(self, url: str, concurrency_semaphore: asyncio.Semaphore) -> AsyncGenerator[CursorPageModel]:
        next_url = f"{url}?{self.query_params}&page[size]=1000&page[cursor]="
        while True:
            async with concurrency_semaphore:
                json = await get_json(next_url)
                page = CursorPageModel.model_validate_json(json)

            yield page

            match page.links.next:
                case None:
                    return
                case str() as next:
                    next_url = next
                case never:
                    assert_never(never)



# SPDX-FileCopyrightText: 2025-present Doug Richardson <git@rekt.email>
# SPDX-License-Identifier: MIT

import logging

import aiofiles
import click

from dogcrud.core.context import config_context
from dogcrud.core.data import resource_type_for_filename

logger = logging.getLogger(__name__)


@click.command()
@click.argument("filename", type=click.Path(exists=True))
def restore(filename: str) -> None:
    """
    Restore a datadog resource from a JSON file in the file system.

    FILENAME is the name of the file to restore.
    """
    config_context().run_in_context(async_restore(filename))


async def async_restore(filename: str) -> None:
    rt = resource_type_for_filename(filename)
    resource_id = rt.resource_id(filename)
    async with aiofiles.open(filename, "rb") as file:
        get_data = await file.read()
    put_data = rt.transform_get_to_put(get_data)
    resource_path = rt.rest_path(resource_id)
    logger.info(f"Restoring {resource_path} from {filename}")
    await rt.put(resource_id, put_data)
    logger.info(f"Restored {resource_path}")



# SPDX-FileCopyrightText: 2025-present Doug Richardson <git@rekt.email>
# SPDX-License-Identifier: MIT

import asyncio
import logging

import click
import orjson

from dogcrud.core.context import config_context
from dogcrud.core.metric_metadata_resource_type import MetricMetadataResourceType
from dogcrud.core.metric_resource_type import MetricResourceType
from dogcrud.core.resource_type import ResourceType
from dogcrud.core.resource_type_registry import resource_types
from dogcrud.core.standard_resource_type import StandardResourceType

logger = logging.getLogger(__name__)


@click.group(name="list")
@click.option(
    "--output",
    "-o",
    type=click.Choice(["table", "json"], case_sensitive=False),
    default="table",
    help="Output format.",
)
@click.pass_context
def list_resources(ctx: click.Context, output: str) -> None:
    """
    List datadog resources without saving them.
    """
    ctx.ensure_object(dict)
    ctx.obj["output_format"] = output


def create_list_command(rt: ResourceType):
    @list_resources.command(
        name=rt.rest_path(),
        help=f"""
    List {rt.rest_path()} resources.
    """,
    )
    @click.pass_context
    def list_resource_command(ctx: click.Context):
        output_format = ctx.obj["output_format"]
        coro = list_all_resources_of_type(rt, output_format)
        config_context().run_in_context(coro)


for rt in resource_types():
    create_list_command(rt)


@list_resources.command(
    name="all",
    help=f"""
List all datadog resources supported by this tool.

Includes: {", ".join(rt.rest_path() for rt in resource_types())}

Note: Disabled resource types are excluded by default. Use --include-disabled to include them.
""",
)
@click.pass_context
def list_all(ctx: click.Context) -> None:
    """
    List all datadog resources.
    """
    output_format = ctx.obj["output_format"]
    config_context().run_in_context(list_all_resource_types(output_format))


async def list_all_resource_types(output_format: str):
    include_disabled = config_context().include_disabled
    async with asyncio.TaskGroup() as tg:
        for resource_type in resource_types():
            if resource_type.disabled and not include_disabled:
                continue
            tg.create_task(list_all_resources_of_type(resource_type, output_format))


async def list_all_resources_of_type(resource_type: ResourceType, output_format: str) -> None:
    prefix = f"list all {resource_type.rest_path()}"
    logger.debug(f"{prefix}: Starting")

    items = []

    # Match specific resource types - order matters!
    match resource_type:
        case MetricMetadataResourceType() | MetricResourceType():
            # Metric types override list_ids and don't support standard pagination
            items = [{"id": resource_id} async for resource_id in resource_type.list_ids()]
        case StandardResourceType():
            # StandardResourceType has pagination_strategy which provides richer item details
            async for page in resource_type.pagination_strategy.pages(
                f"api/{resource_type.rest_path()}", resource_type.concurrency_semaphore
            ):
                items.extend(page.items)
        case _:
            # Fallback for any other resource types
            items = [{"id": resource_id} async for resource_id in resource_type.list_ids()]

    match output_format:
        case "json":
            output_json(resource_type, items)
        case _:
            output_table(resource_type, items)

    logger.debug(f"{prefix}: Listed {len(items)} items.")


def output_json(resource_type: ResourceType, items: list) -> None:
    """Output items as JSON."""
    output = {
        "resource_type": resource_type.rest_path(),
        "items": items,
        "count": len(items),
    }
    click.echo(orjson.dumps(output, option=orjson.OPT_INDENT_2).decode())


def output_table(resource_type: ResourceType, items: list) -> None:
    """Output items as a simple text list."""
    click.echo(f"{resource_type.rest_path()} ({len(items)} resources)")
    for item in items:
        item_id = item.get("id", "N/A")
        # Try to find a human-readable name/title field
        name = (
            item.get("name")
            or item.get("title")
            or item.get("attributes", {}).get("name")
            or item.get("attributes", {}).get("table_name")
        )

        if name:
            click.echo(f"{item_id} {name}")
        else:
            click.echo(f"{item_id}")



# SPDX-FileCopyrightText: 2025-present Doug Richardson <git@rekt.email>
# SPDX-License-Identifier: MIT

import asyncio
import os

import aiofiles
import orjson

from dogcrud.core.resource_type import ResourceType
from dogcrud.core.resource_type_registry import resource_types


def format_json(json: bytes) -> bytes:
    """
    Pretty print JSON with sorted keys so that JSON files are easier to diff.
    """
    parsed_json = orjson.loads(json)
    return orjson.dumps(parsed_json, option=orjson.OPT_INDENT_2 | orjson.OPT_SORT_KEYS | orjson.OPT_APPEND_NEWLINE)


async def write_formatted_json(json: bytes, filename: str) -> None:
    """
    Write JSON to a file, formatted with keys sorted to make diffing files
    easier.
    """
    formatted_json = await asyncio.to_thread(format_json, json)

    async with aiofiles.open(filename, "wb") as out:
        await out.write(formatted_json)


def resource_type_for_filename(filename: str) -> ResourceType:
    filename = os.path.abspath(filename)

    for rt in resource_types():
        if str(rt.local_path()) in filename:
            return rt

    msg = f"No resource type found for {filename}"
    raise RuntimeError(msg)



# SPDX-FileCopyrightText: 2025-present Doug Richardson <git@rekt.email>
# SPDX-License-Identifier: MIT

import asyncio
import logging
import time

import aiohttp.client_exceptions

from dogcrud.core.context import async_run_context

logger = logging.getLogger(__name__)


class DatadogAPIBadRequestError(aiohttp.client_exceptions.ClientResponseError):
    """400 Bad Request error from Datadog API with error body preserved."""

    def __init__(self, request_info, history, status: int, message: str, headers, error_body: str):
        super().__init__(request_info, history, status=status, message=message, headers=headers)
        self.error_body = error_body


async def get_json(path: str) -> bytes:
    """
    Get a Datadog JSON resource. This function is used instead of
    https://datadog-api-client.readthedocs.io/ when we need to download a lot
    of objects and the datadog-api-client API is too slow (the slowness seems
    to be related to the OpenApi parsing on each response). For many cases
    (like backups) we just want to download the RAW json, so no parsing is
    necessary.

    path is the part of the Datadog API URL after https://api.datadoghq.com/.
    For example, to make a request to get all dashboards using
    https://api.datadoghq.com/api/v1/dashboard you'd pass in
    path="v1/dashboard".

    If the request is rate limited with an HTTP 429, this function will sleep
    for X-RateLimit-Reset (a datadog response header) seconds before trying
    again.

    Documentation for Datadog APIs can be found at
    https://docs.datadoghq.com/api/latest/.
    """
    url = path if path.startswith("https://") else f"https://api.datadoghq.com/{path}"
    headers = {"accept": "application/json"}
    ratelimit_reset = 0.0

    retry = 0
    while True:
        async with async_run_context().concurrent_requests_semaphore:
            t0 = time.perf_counter()
            async with async_run_context().datadog_session.get(url, headers=headers) as resp:
                duration = time.perf_counter() - t0
                logger.debug(f"get_json: url={url}, status={resp.status}, duration={duration:.3f}s, retry={retry}")
                match resp.status:
                    case 200:
                        return await resp.read()
                    case 429:
                        # https://docs.datadoghq.com/api/latest/rate-limits/
                        ratelimit_reset = float(resp.headers["X-RateLimit-Reset"])
                    case 400:
                        # Read error body before raising so we can include it in the exception
                        error_body = await resp.text()
                        raise DatadogAPIBadRequestError(
                            request_info=resp.request_info,
                            history=resp.history,
                            status=resp.status,
                            message=resp.reason or "Bad Request",
                            headers=resp.headers,
                            error_body=error_body,
                        )
                    case _:
                        resp.raise_for_status()
                        # above will raise for most HTTP error, but if we get
                        # an unexpected HTTP success code, we will raise
                        # RuntimeError.
                        msg = "Unexpected status code {resp.status} for {url}"
                        raise RuntimeError(msg)

        logger.info(f"Sleeping until rate limit reset in {ratelimit_reset} seconds for {url}")
        await asyncio.sleep(float(ratelimit_reset))
        retry += 1


async def put_json(path: str, body: bytes) -> None:
    url = f"https://api.datadoghq.com/{path}"
    headers = {"content-type": "application/json"}

    async with async_run_context().concurrent_requests_semaphore:
        t0 = time.perf_counter()
        async with async_run_context().datadog_session.put(url, data=body, headers=headers) as resp:
            duration = time.perf_counter() - t0
            msg = f"put_json: url={url}, body={len(body)} bytes, status={resp.status}, duration={duration:.3f}s response={await resp.text()}"
            if resp.ok:
                logger.debug(msg)
            else:
                logger.error(msg)
            resp.raise_for_status()


async def patch_json(path: str, body: bytes) -> None:
    url = f"https://api.datadoghq.com/{path}"
    headers = {"content-type": "application/json"}

    async with async_run_context().concurrent_requests_semaphore:
        t0 = time.perf_counter()
        async with async_run_context().datadog_session.patch(url, data=body, headers=headers) as resp:
            duration = time.perf_counter() - t0
            msg = f"patch_json: url={url}, body={len(body)} bytes, status={resp.status}, duration={duration:.3f}s response={await resp.text()}"
            if resp.ok:
                logger.debug(msg)
            else:
                logger.error(msg)
            resp.raise_for_status()


async def post_json(path: str, body: bytes) -> None:
    url = f"https://api.datadoghq.com/{path}"
    headers = {"content-type": "application/json"}

    async with async_run_context().concurrent_requests_semaphore:
        t0 = time.perf_counter()
        async with async_run_context().datadog_session.post(url, data=body, headers=headers) as resp:
            duration = time.perf_counter() - t0
            msg = f"post_json: url={url}, body={len(body)} bytes, status={resp.status}, duration={duration:.3f}s response={await resp.text()}"
            if resp.ok:
                logger.debug(msg)
            else:
                logger.error(msg)
            resp.raise_for_status()



# SPDX-FileCopyrightText: 2025-present Doug Richardson <git@rekt.email>
# SPDX-License-Identifier: MIT

from collections.abc import Sequence
from functools import partial

from dogcrud.core.metric_metadata_resource_type import MetricMetadataResourceType
from dogcrud.core.metric_resource_type import MetricResourceType
from dogcrud.core.pagination import (
    IDOffsetPagination,
    ItemOffsetPagination,
    NoPagination,
)
from dogcrud.core.reference_table_resource_type import ReferenceTableResourceType
from dogcrud.core.resource_type import ResourceType
from dogcrud.core.standard_resource_type import StandardResourceType
from dogcrud.core.transformers import data_at_key


def resource_types() -> Sequence[ResourceType]:
    """
    The Datadog Resource type definitions that are used to provide common CLI
    implementations for each type.

    https://docs.datadoghq.com/api/latest/
    """
    return (
        StandardResourceType(
            rest_base_path="v1/dashboard",
            webpage_base_path="dashboard",
            max_concurrency=20,
            pagination_strategy=ItemOffsetPagination(offset_query_param="start", items_key="dashboards"),
        ),
        StandardResourceType(
            rest_base_path="v1/monitor",
            webpage_base_path="monitors",
            max_concurrency=100,
            pagination_strategy=IDOffsetPagination(offset_query_param="id_offset"),
        ),
        StandardResourceType(
            rest_base_path="v1/logs/config/pipelines",
            webpage_base_path="logs/pipelines/pipeline/edit",
            max_concurrency=100,
            pagination_strategy=NoPagination(),
        ),
        StandardResourceType(
            rest_base_path="v1/slo",
            webpage_base_path="slo",
            max_concurrency=100,
            pagination_strategy=ItemOffsetPagination(offset_query_param="offset", items_key="data"),
            webpage_suffix="/edit",
            get_to_put_transformer=partial(data_at_key, "data"),
        ),
        StandardResourceType(
            rest_base_path="v2/logs/config/metrics",
            webpage_base_path="logs/pipelines/generate-metrics",
            max_concurrency=100,
            pagination_strategy=NoPagination(items_key="data"),
        ),
        StandardResourceType(
            rest_base_path="v2/workflows",
            webpage_base_path="workflow",
            max_concurrency=100,
            pagination_strategy=NoPagination(items_key="data"),
            disabled=True,
        ),
        ReferenceTableResourceType(
            max_concurrency=100,
        ),
        MetricMetadataResourceType(
            max_concurrency=100,
        ),
        MetricResourceType(
            max_concurrency=100,
        ),
    )



# SPDX-FileCopyrightText: 2025-present Doug Richardson <git@rekt.email>
# SPDX-License-Identifier: MIT

import asyncio
import logging
from collections.abc import AsyncGenerator
from typing import override

from pydantic import BaseModel

from dogcrud.core.metrics import list_metrics
from dogcrud.core.pagination import CursorDataItemModel, NoPagination
from dogcrud.core.resource_type import IDType
from dogcrud.core.standard_resource_type import StandardResourceType

logger = logging.getLogger(__name__)


class MetricMetadataModel(BaseModel):
    description: str | None
    integration: str | None
    per_unit: str | None
    short_name: str
    statsd_interval: int
    type: str
    unit: str


class MetricMetadataResourceType(StandardResourceType):
    """
    A Datadog metric resource. This resource doesn't fit in with a StandardResourceType since
    - you have an idea of "active" metrics that the initial list is populated from from a time range
    - you get a list of metrics with a subset of metadata
    - there is no get ID list and get specific resource json like StandardResourceType
    """

    def __init__(
        self,
        max_concurrency: int,
    ) -> None:
        super().__init__(
            rest_base_path="v1/metrics",
            webpage_base_path="metric/summary?metric=",
            max_concurrency=max_concurrency,
            pagination_strategy=NoPagination(),
        )
        self.rest_base_path = "v1/metrics"
        self.concurrency_semaphore = asyncio.BoundedSemaphore(max_concurrency)
        self._id_to_item_index: dict[IDType, CursorDataItemModel] = {}

    @override
    async def list_ids(self) -> AsyncGenerator[IDType]:
        async for metric in list_metrics(self.concurrency_semaphore):
            yield metric.id

    @override
    def webpage_url(self, resource_id: IDType) -> str:
        return f"https://app.datadoghq.com/{self.webpage_base_path}{resource_id}"



import orjson

from dogcrud.core import data


def test_ends_in_newline():
    assert data.format_json(b'{"one": 123}')[-1] == ord("\n")


def test_keys_sorted():
    json = data.format_json(b'{"b": 1, "z": 1, "a": { "subb": 2, "subz": 3, "suba": 4}}')
    parsed_json = orjson.loads(json)

    # Performing list(d) on a dictionary returns a list of all the keys used in the dictionary, in insertion order
    # https://docs.python.org/3/tutorial/datastructures.html#dictionaries
    assert list(parsed_json) == ["a", "b", "z"]

    assert list(parsed_json["a"]) == ["suba", "subb", "subz"]



# SPDX-FileCopyrightText: 2025-present Doug Richardson <git@rekt.email>
# SPDX-License-Identifier: MIT

import logging
import pathlib
from typing import override

import orjson

from dogcrud.core.pagination import LimitOffsetPagination
from dogcrud.core.resource_type import IDType
from dogcrud.core.standard_resource_type import StandardResourceType

logger = logging.getLogger(__name__)


class ReferenceTableResourceType(StandardResourceType):
    """
    A Datadog reference table resource. This overrides webpage_url to use the table_name
    from the resource attributes instead of the UUID ID.
    """

    def __init__(
        self,
        max_concurrency: int,
    ) -> None:
        super().__init__(
            rest_base_path="v2/reference-tables/tables",
            webpage_base_path="reference-tables",
            max_concurrency=max_concurrency,
            pagination_strategy=LimitOffsetPagination(
                limit=100,
                limit_query_param="page[limit]",
                offset_query_param="page[offset]",
                items_key="data",
            ),
        )

    @override
    def webpage_url(self, resource_id: IDType) -> str:
        saved_file = pathlib.Path(f"saved/{self.rest_base_path}/{resource_id}.json")
        data = orjson.loads(saved_file.read_bytes())
        table_name = data["data"]["attributes"]["table_name"]
        return f"https://app.datadoghq.com/{self.webpage_base_path}/{table_name}"



# SPDX-FileCopyrightText: 2025-present Doug Richardson <git@rekt.email>
# SPDX-License-Identifier: MIT

from collections.abc import AsyncGenerator
from pathlib import Path
from typing import Protocol

type IDType = int | str


class ResourceType(Protocol):
    """
    A Datadog resource type that can have multiple instances, like dashboards,
    monitors, and metrics.
    """

    disabled: bool

    def rest_path(self, resource_id: IDType | None = None) -> str: ...
    def local_path(self, resource_id: IDType | None = None) -> Path: ...
    async def get(self, resource_id: IDType) -> bytes: ...
    async def put(self, resource_id: IDType, data: bytes) -> None: ...
    def transform_get_to_put(self, data: bytes) -> bytes: ...
    def list_ids(self) -> AsyncGenerator[IDType]: ...
    async def read_local_json(self, resource_id: IDType) -> bytes: ...
    def webpage_url(self, resource_id: IDType) -> str: ...
    def resource_id(self, filename: str) -> IDType: ...



from collections.abc import AsyncGenerator

from dogcrud.core.pagination import CursorDataItemModel, CursorPagination


async def list_metrics(semaphore) -> AsyncGenerator[CursorDataItemModel]:
    # data[i].attributes is only set when filter[configured]=true is set,
    # so do 2 paginations—one with false and one with true—in order to get
    # the attributes.
    for configured in ("false", "true"):
        paginator = CursorPagination(query_params=f"filter[configured]={configured}")
        async for page in paginator.pages("api/v2/metrics", semaphore):
            for item in page.data:
                yield item



from collections.abc import Callable

import orjson

type GetToPut = Callable[[bytes], bytes]


def identity(data: bytes) -> bytes:
    """
    Use identity transformer when GET data is same as PUT data.
    """
    return data


def data_at_key(key: str, data: bytes) -> bytes:
    """
    Use data_at_key trasnformer when GET data is a dictionary that contains the PUT data at a key.
    """
    json = orjson.loads(data)
    inner_data = json[key]
    return orjson.dumps(inner_data)
